@@ -1,16 +1,11 @@
-/* -*- indent-tabs-mode: nil; js-indent-level: 2 -*- */
-/* vim: set sw=2 ts=2 et lcs=trail\:.,tab\:>~ : */
 /* This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
-/**
+/*
  * nsILoginManagerStorage implementation for the JSON back-end.
  */
 
-////////////////////////////////////////////////////////////////////////////////
-//// Globals
-
 "use strict";
 
 const { classes: Cc, interfaces: Ci, utils: Cu, results: Cr } = Components;
@@ -32,16 +27,13 @@ XPCOMUtils.defineLazyServiceGetter(this, "gUUIDGenerator",
                                    "@mozilla.org/uuid-generator;1",
                                    "nsIUUIDGenerator");
 
-////////////////////////////////////////////////////////////////////////////////
-//// LoginManagerStorage_json
-
-this.LoginManagerStorage_json = function () {}
+this.LoginManagerStorage_json = function () {};
 
 this.LoginManagerStorage_json.prototype = {
-  classID : Components.ID("{c00c432d-a0c9-46d7-bef6-9c45b4d07341}"),
-  QueryInterface : XPCOMUtils.generateQI([Ci.nsILoginManagerStorage]),
+  classID: Components.ID("{c00c432d-a0c9-46d7-bef6-9c45b4d07341}"),
+  QueryInterface: XPCOMUtils.generateQI([Ci.nsILoginManagerStorage]),
 
-  __crypto : null,  // nsILoginManagerCrypto service
+  __crypto: null,  // nsILoginManagerCrypto service
   get _crypto() {
     if (!this.__crypto)
       this.__crypto = Cc["@mozilla.org/login-manager/crypto/SDR;1"].
@@ -49,27 +41,7 @@ this.LoginManagerStorage_json.prototype = {
     return this.__crypto;
   },
 
-  /*
-   * log
-   *
-   * Internal function for logging debug messages to the Error Console.
-   */
-  log : function (message) {
-    if (!this._debug)
-      return;
-    dump("PwMgr json: " + message + "\n");
-    Services.console.logStringMessage("PwMgr json: " + message);
-  },
-  _debug : false,
-
-
-  /*
-   * initialize
-   *
-   */
-  initialize : function () {
-    this._debug = Services.prefs.getBoolPref("signon.debug");
-
+  initialize() {
     try {
       // Force initialization of the crypto module.
       // See bug 717490 comment 17.
@@ -80,9 +52,9 @@ this.LoginManagerStorage_json.prototype = {
                                   "logins.json");
       this._store = new LoginStore(jsonPath);
 
-      return Task.spawn(function () {
+      return Task.spawn(function* () {
         // Load the data asynchronously.
-        this.log("Opening database at " + this._store.path);
+        this.log("Opening database at", this._store.path);
         yield this._store.load();
 
         // The import from previous versions operates the first time
@@ -114,29 +86,21 @@ this.LoginManagerStorage_json.prototype = {
         Services.prefs.setBoolPref("signon.importedFromSqlite", true);
       }.bind(this)).catch(Cu.reportError);
     } catch (e) {
-      this.log("Initialization failed: " + e);
+      this.log("Initialization failed:", e);
       throw new Error("Initialization failed");
     }
   },
 
-
-  /*
-   * terminate
-   *
+  /**
    * Internal method used by regression tests only.  It is called before
    * replacing this storage module with a new instance.
    */
-  terminate : function () {
+  terminate() {
     this._store._saver.disarm();
     return this._store.save();
   },
 
-
-  /*
-   * addLogin
-   *
-   */
-  addLogin : function (login) {
+  addLogin(login) {
     this._store.ensureDataReady();
 
     // Throws if there are bogus values.
@@ -189,12 +153,7 @@ this.LoginManagerStorage_json.prototype = {
     this._sendNotification("addLogin", loginClone);
   },
 
-
-  /*
-   * removeLogin
-   *
-   */
-  removeLogin : function (login) {
+  removeLogin(login) {
     this._store.ensureDataReady();
 
     let [idToDelete, storedLogin] = this._getIdForLogin(login);
@@ -210,12 +169,7 @@ this.LoginManagerStorage_json.prototype = {
     this._sendNotification("removeLogin", storedLogin);
   },
 
-
-  /*
-   * modifyLogin
-   *
-   */
-  modifyLogin : function (oldLogin, newLoginData) {
+  modifyLogin(oldLogin, newLoginData) {
     this._store.ensureDataReady();
 
     let [idToModify, oldStoredLogin] = this._getIdForLogin(oldLogin);
@@ -267,34 +221,28 @@ this.LoginManagerStorage_json.prototype = {
     this._sendNotification("modifyLogin", [oldStoredLogin, newLogin]);
   },
 
-
-  /*
-   * getAllLogins
-   *
-   * Returns an array of nsILoginInfo.
+  /**
+   * @return {nsILoginInfo[]}
    */
-  getAllLogins : function (count) {
+  getAllLogins(count) {
     let [logins, ids] = this._searchLogins({});
 
     // decrypt entries for caller.
     logins = this._decryptLogins(logins);
 
-    this.log("_getAllLogins: returning " + logins.length + " logins.");
+    this.log("_getAllLogins: returning", logins.length, "logins.");
     if (count)
       count.value = logins.length; // needed for XPCOM
     return logins;
   },
 
-
-  /*
-   * searchLogins
-   *
+  /**
    * Public wrapper around _searchLogins to convert the nsIPropertyBag to a
    * JavaScript object and decrypt the results.
    *
-   * Returns an array of decrypted nsILoginInfo.
+   * @return {nsILoginInfo[]} which are decrypted.
    */
-  searchLogins : function(count, matchData) {
+  searchLogins(count, matchData) {
     let realMatchData = {};
     // Convert nsIPropertyBag to normal JS object
     let propEnum = matchData.enumerator;
@@ -312,10 +260,7 @@ this.LoginManagerStorage_json.prototype = {
     return logins;
   },
 
-
-  /*
-   * _searchLogins
-   *
+  /**
    * Private method to perform arbitrary searches on any field. Decryption is
    * left to the caller.
    *
@@ -323,12 +268,17 @@ this.LoginManagerStorage_json.prototype = {
    * is an array of encrypted nsLoginInfo and ids is an array of associated
    * ids in the database.
    */
-  _searchLogins : function (matchData) {
+  _searchLogins(matchData) {
     this._store.ensureDataReady();
 
     let conditions = [];
 
     function match(aLogin) {
+      let returnValue = {
+        match: false,
+        strictMatch: true
+      };
+
       for (let field in matchData) {
         let value = matchData[field];
         switch (field) {
@@ -336,7 +286,28 @@ this.LoginManagerStorage_json.prototype = {
           case "formSubmitURL":
             if (value != null) {
               if (aLogin.formSubmitURL != "" && aLogin.formSubmitURL != value) {
-                return false;
+                // Check for cases that don't have fallback matches.
+                if (value == "" || value == "javascript:" ||
+                    aLogin.formSubmitURL == "javascript:" ||
+                    aLogin.formSubmitURL == null) {
+                  return returnValue;
+                }
+
+                // Check if it matches with a different scheme.
+                let loginURI = Services.io.newURI(aLogin.formSubmitURL, null, null);
+                let matchURI = Services.io.newURI(value, null, null);
+
+                if (loginURI.hostPort != matchURI.hostPort) {
+                  return returnValue; // not a match at all
+                }
+
+                if ((loginURI.scheme != "http" && loginURI.scheme != "https") ||
+                    (matchURI.scheme != "http" && matchURI.scheme != "https")) {
+                  // Not a match at all since we only fallback HTTP <=> HTTPS.
+                  return returnValue;
+                }
+
+                returnValue.strictMatch = false; // not a strict match
               }
               break;
             }
@@ -355,9 +326,9 @@ this.LoginManagerStorage_json.prototype = {
           case "timePasswordChanged":
           case "timesUsed":
             if (value == null && aLogin[field]) {
-              return false;
+              return returnValue;
             } else if (aLogin[field] != value) {
-              return false;
+              return returnValue;
             }
             break;
           // Fail if caller requests an unknown property.
@@ -365,12 +336,14 @@ this.LoginManagerStorage_json.prototype = {
             throw new Error("Unexpected field: " + field);
         }
       }
-      return true;
+      returnValue.match = true;
+      return returnValue;
     }
 
-    let foundLogins = [], foundIds = [];
+    let foundLogins = [], foundIds = [], fallbackLogins = [], fallbackIds = [];
     for (let loginItem of this._store.data.logins) {
-      if (match(loginItem)) {
+      let result = match(loginItem);
+      if (result.match) {
         // Create the new nsLoginInfo object, push to array
         let login = Cc["@mozilla.org/login-manager/loginInfo;1"].
                     createInstance(Ci.nsILoginInfo);
@@ -385,23 +358,31 @@ this.LoginManagerStorage_json.prototype = {
         login.timeLastUsed = loginItem.timeLastUsed;
         login.timePasswordChanged = loginItem.timePasswordChanged;
         login.timesUsed = loginItem.timesUsed;
-        foundLogins.push(login);
-        foundIds.push(loginItem.id);
+        // If protocol does not match, use as a fallback login
+        if (result.strictMatch) {
+          foundLogins.push(login);
+          foundIds.push(loginItem.id);
+        } else {
+          fallbackLogins.push(login);
+          fallbackIds.push(loginItem.id);
+        }
       }
     }
 
-    this.log("_searchLogins: returning " + foundLogins.length + " logins");
+    if (!foundLogins.length && fallbackLogins.length) {
+      this.log("_searchLogins: returning", fallbackLogins.length, "fallback logins");
+      return [fallbackLogins, fallbackIds];
+    }
+    this.log("_searchLogins: returning", foundLogins.length, "logins");
     return [foundLogins, foundIds];
   },
 
-  /*
-   * removeAllLogins
-   *
+  /**
    * Removes all logins from storage.
    *
    * Disabled hosts are kept, as one presumably doesn't want to erase those.
    */
-  removeAllLogins : function () {
+  removeAllLogins() {
     this._store.ensureDataReady();
 
     this.log("Removing all logins");
@@ -411,46 +392,31 @@ this.LoginManagerStorage_json.prototype = {
     this._sendNotification("removeAllLogins", null);
   },
 
-
-  /*
-   * getAllDisabledHosts
-   *
-   */
-  getAllDisabledHosts : function (count) {
+  getAllDisabledHosts(count) {
     this._store.ensureDataReady();
 
     let disabledHosts = this._store.data.disabledHosts.slice(0);
 
-    this.log("_getAllDisabledHosts: returning " + disabledHosts.length + " disabled hosts.");
+    this.log("_getAllDisabledHosts: returning", disabledHosts.length, "disabled hosts.");
     if (count)
       count.value = disabledHosts.length; // needed for XPCOM
     return disabledHosts;
   },
 
-
-  /*
-   * getLoginSavingEnabled
-   *
-   */
-  getLoginSavingEnabled : function (hostname) {
+  getLoginSavingEnabled(hostname) {
     this._store.ensureDataReady();
 
-    this.log("Getting login saving is enabled for " + hostname);
+    this.log("Getting login saving is enabled for", hostname);
     return this._store.data.disabledHosts.indexOf(hostname) == -1;
   },
 
-
-  /*
-   * setLoginSavingEnabled
-   *
-   */
-  setLoginSavingEnabled : function (hostname, enabled) {
+  setLoginSavingEnabled(hostname, enabled) {
     this._store.ensureDataReady();
 
     // Throws if there are bogus values.
     LoginHelper.checkHostnameValue(hostname);
 
-    this.log("Setting login saving enabled for " + hostname + " to " + enabled);
+    this.log("Setting login saving enabled for", hostname, "to", enabled);
     let foundIndex = this._store.data.disabledHosts.indexOf(hostname);
     if (enabled) {
       if (foundIndex != -1) {
@@ -467,19 +433,14 @@ this.LoginManagerStorage_json.prototype = {
     this._sendNotification(enabled ? "hostSavingEnabled" : "hostSavingDisabled", hostname);
   },
 
-
-  /*
-   * findLogins
-   *
-   */
-  findLogins : function (count, hostname, formSubmitURL, httpRealm) {
+  findLogins(count, hostname, formSubmitURL, httpRealm) {
     let loginData = {
       hostname: hostname,
       formSubmitURL: formSubmitURL,
       httpRealm: httpRealm
     };
     let matchData = { };
-    for each (let field in ["hostname", "formSubmitURL", "httpRealm"])
+    for (let field of ["hostname", "formSubmitURL", "httpRealm"])
       if (loginData[field] != '')
         matchData[field] = loginData[field];
     let [logins, ids] = this._searchLogins(matchData);
@@ -487,17 +448,12 @@ this.LoginManagerStorage_json.prototype = {
     // Decrypt entries found for the caller.
     logins = this._decryptLogins(logins);
 
-    this.log("_findLogins: returning " + logins.length + " logins");
+    this.log("_findLogins: returning", logins.length, "logins");
     count.value = logins.length; // needed for XPCOM
     return logins;
   },
 
-
-  /*
-   * countLogins
-   *
-   */
-  countLogins : function (hostname, formSubmitURL, httpRealm) {
+  countLogins(hostname, formSubmitURL, httpRealm) {
     let count = {};
     let loginData = {
       hostname: hostname,
@@ -505,38 +461,27 @@ this.LoginManagerStorage_json.prototype = {
       httpRealm: httpRealm
     };
     let matchData = { };
-    for each (let field in ["hostname", "formSubmitURL", "httpRealm"])
+    for (let field of ["hostname", "formSubmitURL", "httpRealm"])
       if (loginData[field] != '')
         matchData[field] = loginData[field];
     let [logins, ids] = this._searchLogins(matchData);
 
-    this.log("_countLogins: counted logins: " + logins.length);
+    this.log("_countLogins: counted logins:", logins.length);
     return logins.length;
   },
 
-
-  /*
-   * uiBusy
-   */
   get uiBusy() {
     return this._crypto.uiBusy;
   },
 
-
-  /*
-   * isLoggedIn
-   */
   get isLoggedIn() {
     return this._crypto.isLoggedIn;
   },
 
-
-  /*
-   * _sendNotification
-   *
+  /**
    * Send a notification when stored data is changed.
    */
-  _sendNotification : function (changeType, data) {
+  _sendNotification(changeType, data) {
     let dataObject = data;
     // Can't pass a raw JS string or array though notifyObservers(). :-(
     if (data instanceof Array) {
@@ -552,17 +497,14 @@ this.LoginManagerStorage_json.prototype = {
     Services.obs.notifyObservers(dataObject, "passwordmgr-storage-changed", changeType);
   },
 
-
-  /*
-   * _getIdForLogin
-   *
+  /**
    * Returns an array with two items: [id, login]. If the login was not
    * found, both items will be null. The returned login contains the actual
    * stored login (useful for looking at the actual nsILoginMetaInfo values).
    */
-  _getIdForLogin : function (login) {
+  _getIdForLogin(login) {
     let matchData = { };
-    for each (let field in ["hostname", "formSubmitURL", "httpRealm"])
+    for (let field of ["hostname", "formSubmitURL", "httpRealm"])
       if (login[field] != '')
         matchData[field] = login[field];
     let [logins, ids] = this._searchLogins(matchData);
@@ -589,26 +531,20 @@ this.LoginManagerStorage_json.prototype = {
     return [id, foundLogin];
   },
 
-
-  /*
-   * _isGuidUnique
-   *
+  /**
    * Checks to see if the specified GUID already exists.
    */
-  _isGuidUnique : function (guid) {
+  _isGuidUnique(guid) {
     this._store.ensureDataReady();
 
     return this._store.data.logins.every(l => l.guid != guid);
   },
 
-
-  /*
-   * _encryptLogin
-   *
+  /**
    * Returns the encrypted username, password, and encrypton type for the specified
    * login. Can throw if the user cancels a master password entry.
    */
-  _encryptLogin : function (login) {
+  _encryptLogin(login) {
     let encUsername = this._crypto.encrypt(login.username);
     let encPassword = this._crypto.encrypt(login.password);
     let encType     = this._crypto.defaultEncType;
@@ -616,10 +552,7 @@ this.LoginManagerStorage_json.prototype = {
     return [encUsername, encPassword, encType];
   },
 
-
-  /*
-   * _decryptLogins
-   *
+  /**
    * Decrypts username and password fields in the provided array of
    * logins.
    *
@@ -630,10 +563,10 @@ this.LoginManagerStorage_json.prototype = {
    * to lose unencrypted entries (eg, because the user clicked Cancel
    * instead of entering their master password)
    */
-  _decryptLogins : function (logins) {
+  _decryptLogins(logins) {
     let result = [];
 
-    for each (let login in logins) {
+    for (let login of logins) {
       try {
         login.username = this._crypto.decrypt(login.username);
         login.password = this._crypto.decrypt(login.password);
@@ -651,4 +584,9 @@ this.LoginManagerStorage_json.prototype = {
   },
 };
 
+XPCOMUtils.defineLazyGetter(this.LoginManagerStorage_json.prototype, "log", () => {
+  let logger = LoginHelper.createLogger("Login storage");
+  return logger.log.bind(logger);
+});
+
 this.NSGetFactory = XPCOMUtils.generateNSGetFactory([LoginManagerStorage_json]);
